<?php

namespace THM\DownloadExplorer\Utility;

/***************************************************************
 *
 *  Copyright notice
 *
 *  (c) 2014-2019
 *           Jan Haffner <info@jan-haffner.de>
 *           Rafael Kähm <Rafael@Kaehms.de>
 *           Samuel Mehrbrodt <samuel.mehrbrodt@mni.thm.de>
 *
 *  All rights reserved
 *
 *  This script is part of the TYPO3 project. The TYPO3 project is
 *  free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  The GNU General Public License can be found at
 *  http://www.gnu.org/copyleft/gpl.html.
 *
 *  This script is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  This copyright notice MUST APPEAR in all copies of the script!
 ***************************************************************/

use \TYPO3\CMS\Core\Utility\GeneralUtility;

class Files extends GeneralUtility
{

    /**
     * Settings for file explorer.
     *
     * @var array
     */
    protected static $settings = array();

    /**
     * Set settings.
     *
     * @param array $settings
     */
    public static function setSettings(array $settings)
    {
        self::$settings = $settings;
        if (self::$settings['hideEmptyFolders'] == 'typoscript') {
            self::$settings['hideEmptyFolders'] = self::$settings['hideEmptyFoldersDefault'];
        }
        if (self::$settings['hideRootFolder'] == 'typoscript') {
            self::$settings['hideRootFolder'] = self::$settings['hideRootFolderDefault'];
        }
        self::$settings['forbiddenFiles'] = array_map('trim', explode(",", self::$settings['forbiddenFiles']));
        self::$settings['excludedExtensions'] = array_map('trim', explode(",", self::$settings['excludedExtensions']));
        self::$settings['excludedExtensions'] = array_merge(self::$settings['forbiddenFiles'], self::$settings['excludedExtensions']);
    }

    /**
     * Read a directory and return a list with files and folders in it.
     *
     * @param string $basePath        The directory to list
     * @param array  $filesAndFolders The array which is returned
     * @param bool   $followSymLinks  Whether to follow symbolic links
     * @param int    $foldersDepth    How deep should I walk?
     *
     * @throws \Exception When $basePath is not a directory
     *
     * @return array Key: Filename; Value: ExtendedSplFileInfo
     */
    public static function readDirectoryRecursively($basePath, &$filesAndFolders = array(), $followSymLinks = true,
                                                    $foldersDepth = 99)
    {
        $resourceFactory = \TYPO3\CMS\Core\Resource\ResourceFactory::getInstance();
        $basePath = $resourceFactory->getFolderObjectFromCombinedIdentifier($basePath)->getPublicUrl();

        if (!is_dir($basePath)) {
            throw new \Exception('"' . $basePath . '" is not a directory.', 1408394305);
        }

        $directoryIterator = new \FilesystemIterator($basePath, $followSymLinks ? \FilesystemIterator::FOLLOW_SYMLINKS : null);
        $directoryIterator->setInfoClass('THM\DownloadExplorer\Utility\ExtendedSplFileInfo');

        foreach ($directoryIterator as $fileInfo) {
            if (self::isFileOrDirExcluded($fileInfo)) {
                continue;
            }
            // Check access rights in first level
            if (self::$settings['foldersDepth'] == $foldersDepth  &&
                $fileInfo->isDir() && !self::canViewFolder($fileInfo)) {
                continue;
            }
            if ($fileInfo->isDir() && $foldersDepth >= 1) {
                $subFoldersAndFiles = array();
                self::readDirectoryRecursively($fileInfo->getPathname(), $subFoldersAndFiles, $followSymLinks,
                                               $foldersDepth - 1);

                $fileInfo->setSubFoldersAndFiles($subFoldersAndFiles);
                if (self::isDirEmpty($fileInfo)) {
                    continue;
                }
            }
            // #11555 skip folders on the last level
            if ($fileInfo->isDir() && $foldersDepth < 1) {
                continue;
            }
            $filesAndFolders[$fileInfo->getFilename()] = $fileInfo;
        }

        self::orderFilesAndFoldersRecursively($filesAndFolders, self::$settings['sorting'], self::$settings['sortingascdesc']);

        return $filesAndFolders;
    }

    /**
     * Checks if file should be excluded.
     *
     * Calls all available filter methods for checking file or dir visibility;
     *
     * @param ExtendedSplFileInfo $fileInfo
     *
     * @return bool
     */
    private static function isFileOrDirExcluded(ExtendedSplFileInfo $fileInfo)
    {
        return self::isFileOrDirHidden($fileInfo)
            || ($fileInfo->isFile() && self::isFileExcludedByExtension($fileInfo))
            || self::isUsersExclusionPatternsForFilesOrDirsMatched($fileInfo);
    }

    /**
     * Determines whether the currently logged in FE user is allowed to view the given folder.
     *
     * @param ExtendedSplFileInfo $fileInfo
     *
     * @return bool
     */
    private static function canViewFolder(ExtendedSplFileInfo $fileInfo)
    {
        return self::$settings['grouprights'] == 0 ||
               self::frontendUserHasRole($fileInfo->getFilename());
    }

    /**
     * Determines whether the currently logged in FE user belongs to the specified usergroup.
     *
     * @param string $role The usergroup (either the usergroup uid or its title)
     *
     * @return bool TRUE if the currently logged in FE user belongs to $role
     */
    protected function frontendUserHasRole($role)
    {
        if (!isset($GLOBALS['TSFE']) || !$GLOBALS['TSFE']->loginUser) {
            return false;
        }
        if (is_numeric($role)) {
            return is_array($GLOBALS['TSFE']->fe_user->groupData['uid']) && in_array($role, $GLOBALS['TSFE']->fe_user->groupData['uid']);
        } else {
            return is_array($GLOBALS['TSFE']->fe_user->groupData['title']) && in_array($role, $GLOBALS['TSFE']->fe_user->groupData['title']);
        }
    }

    /**
     * Checks if file or folder is UNIX Like hidden: if file/directory name begins with dot.
     *
     * @param ExtendedSplFileInfo $fileInfo
     *
     * @return bool
     */
    private static function isFileOrDirHidden(ExtendedSplFileInfo $fileInfo)
    {
        return preg_match('/^\./', $fileInfo->getFilename());
    }

    /**
     * Check if file type (extension) is excluded.
     *
     * @param ExtendedSplFileInfo $fileInfo
     *
     * @return bool
     */
    private static function isFileExcludedByExtension(ExtendedSplFileInfo $fileInfo)
    {
        foreach (self::$settings['excludedExtensions'] as $ext) {
            if ($fileInfo->getExtension() ==  $ext) {
                return true;
            }
        }

        return false;
    }

    /**
     * Check regex for files or folders.
     *
     * @param ExtendedSplFileInfo $fileInfo
     *
     * @return bool
     */
    private static function isUsersExclusionPatternsForFilesOrDirsMatched(ExtendedSplFileInfo $fileInfo)
    {
        if ($fileInfo->isFile()) {
            $regExpattern = self::$settings['regexFiles'];
        } elseif ($fileInfo->isDir()) {
            $regExpattern = self::$settings['regexFolders'];
        }
        if (!empty($regExpattern) && preg_match($regExpattern, $fileInfo->getFilename())) {
            return true;
        }

        return false;
    }

    /**
     * Checks if given dir is empty dir.
     *
     * @param ExtendedSplFileInfo $fileInfo
     *
     * @return bool
     */
    private static function isDirEmpty(ExtendedSplFileInfo $fileInfo)
    {
        $filesAndDirs = $fileInfo->getSubFoldersAndFiles();

        return self::$settings['hideEmptyFolders'] && empty($filesAndDirs);
    }

    /**
     * Orders files and folders.
     *
     * @param array  $filesAndFolders
     * @param string $orderBy
     * @param string $ascDesc
     * @param int    $foldersDepth
     *
     * @return array with sorted files and folders
     */
    public static function orderFilesAndFoldersRecursively(array &$filesAndFolders, $orderBy = 'name', $ascDesc = 'asc', $foldersDepth = 99)
    {
        foreach ($filesAndFolders as $fileOrFolder) {
            /* @var ExtendedSplFileInfo $fileOrFolder */
            if ($fileOrFolder->isDir()) {
                self::orderFilesAndFoldersRecursively($fileOrFolder->getSubFoldersAndFiles(), $orderBy, $ascDesc, $foldersDepth - 1);
            }
        }

        switch ($orderBy) {
            case 'ctime':
                usort($filesAndFolders, function ($a, $b) use ($ascDesc) {
                    /* @var ExtendedSplFileInfo $a */ /* @var ExtendedSplFileInfo $b */
                    if ($a->getCTime() > $b->getCTime()) {
                        return $ascDesc == 'asc' ? 1 : -1;
                    }
                    if ($a->getCTime() < $b->getCTime()) {
                        return $ascDesc == 'asc' ? -1 : 1;
                    }

                    return 0;
                });
                break;
            case 'type':
            case 'extension':
                usort($filesAndFolders, function ($a, $b) use ($ascDesc) {
                    /* @var ExtendedSplFileInfo $a */ /* @var ExtendedSplFileInfo $b */
                    if ($ascDesc == 'asc') {
                        return strcasecmp($a->getExtension(), $b->getExtension());
                    }

                    return strcasecmp($a->getExtension(), $b->getExtension()) * -1;
                });
                break;
            case 'mtime':
                usort($filesAndFolders, function ($a, $b) use ($ascDesc) {
                    /* @var ExtendedSplFileInfo $a */ /* @var ExtendedSplFileInfo $b */
                    if ($a->getMTime() > $b->getMTime()) {
                        return $ascDesc == 'asc' ? 1 : -1;
                    }
                    if ($a->getMTime() < $b->getMTime()) {
                        return $ascDesc == 'asc' ? -1 : 1;
                    }

                    return 0;
                });
                break;
            case 'size':
                usort($filesAndFolders, function ($a, $b) use ($ascDesc) {
                    /* @var ExtendedSplFileInfo $a */ /* @var ExtendedSplFileInfo $b */
                    if ($a->getSize() > $b->getSize()) {
                        return $ascDesc == 'asc' ? 1 : -1;
                    }
                    if ($a->getSize() < $b->getSize()) {
                        return $ascDesc == 'asc' ? -1 : 1;
                    }

                    return 0;
                });
                break;
            case 'name':
            case 'filename':
            case 'default':
            default:
                usort($filesAndFolders, function ($a, $b) use ($ascDesc) {
                    /* @var ExtendedSplFileInfo $a */ /* @var ExtendedSplFileInfo $b */
                    if ($ascDesc == 'asc') {
                        return strcasecmp($a->getFilename(), $b->getFilename());
                    }

                    return strcasecmp($a->getFilename(), $b->getFilename()) * -1;
                });
        }

        return $filesAndFolders;
    }
}
